// Copyright 1997 DJ Delorie <dj@delorie.com>
// Permission granted to do whatever you want with these sources,
// provided they are not changed

#include <stdio.h>
#include <string.h>

FILE *outf = 0;
int verbose = 0;
int quiet = 0;

int old_percent = -1;
int count = 0;

char wins[][3][2] = {
  {{0,0},{0,1},{0,2}},
  {{1,0},{1,1},{1,2}},
  {{2,0},{2,1},{2,2}},
  {{0,0},{1,0},{2,0}},
  {{0,1},{1,1},{2,1}},
  {{0,2},{1,2},{2,2}},
  {{0,0},{1,1},{2,2}},
  {{0,2},{1,1},{2,0}}
};
#define NUM_WINS (sizeof(wins)/sizeof(wins[0]))

#define dot '-'

char saved_results[524288];
char board[3][3] = {{dot,dot,dot},{dot,dot,dot},{dot,dot,dot}};

#define NO_MOVE	0
#define X_WINS	1
#define O_WINS	2
#define DRAW	3
#define UNKNOWN	4

char *describe[] = {
  "No Move",
  "1 Wins",
  "0 Wins",
  "Draw"
};

char key[] = "-10d";

int
board_to_int(char this_player)
{
  int i, rv=0;
  for (i=0; i<9; i++)
  {
    int x = i/3;
    int y = i%3;
    if (board[x][y] == '1') rv |= 1<<(i*2);
    if (board[x][y] == '0') rv |= 2<<(i*2);
  }
  if (this_player == '1') rv |= 1<<18;
  return rv;
}

void
note(char player, char *move)
{
  if (outf)
    fprintf(outf, "%c%c%c%c%c%c%c%c%c%c %s\n",
	    board[0][0], board[0][1], board[0][2],
	    board[1][0], board[1][1], board[1][2],
	    board[2][0], board[2][1], board[2][2],
	    player, move);

  if (!quiet && !verbose)
  {
    count++;
    int new_percent = count * 100 / 39366;
    if (new_percent != old_percent)
    {
      printf("\r%3d%% ", new_percent);
      old_percent = new_percent;
      fflush(stdout);
    }
  }
}
void
note(char player, char move)
{
  char m[2];
  m[0] = move;
  m[1] = 0;
  note(player, m);
}

int
figure_board(char this_player, char other_player, int cells_filled)
{
  int i, x, y;

  int board_index = board_to_int(this_player);
  if (saved_results[board_index] != UNKNOWN)
    return saved_results[board_index];

  // First, see if the configuration has a winner
  for (i=0; i<(int)NUM_WINS; i++)
  {
    if (board[wins[i][0][0]][wins[i][0][1]] == '1'
	&& board[wins[i][1][0]][wins[i][1][1]] == '1'
	&& board[wins[i][2][0]][wins[i][2][1]] == '1')
    {
      saved_results[board_index] = X_WINS;
      note(this_player, "1 wins");
      return X_WINS;
    }
    if (board[wins[i][0][0]][wins[i][0][1]] == '0'
	&& board[wins[i][1][0]][wins[i][1][1]] == '0'
	&& board[wins[i][2][0]][wins[i][2][1]] == '0')
    {
      saved_results[board_index] = O_WINS;
      note(this_player, "0 wins");
      return O_WINS;
    }
  }

  // See if it's a draw
  if (cells_filled == 9)
  {
    saved_results[board_index] = DRAW;
    note(this_player, "Draw");
    return DRAW;
  }

  // Now, try other moves.

  int results[9] = {0,0,0,0,0,0,0,0,0}; // NO_MOVE
  int moves[9] = {0,0,0,0,0,0,0,0,0};

  for (i=0; i<9; i++)
  {
    x = i/3;
    y = i%3;

    if (board[x][y] != dot)
      continue;

    board[x][y] = this_player;
    results[i] = figure_board(other_player, this_player, cells_filled+1);
    moves[i] = results[i] >> 4;
    results[i] &= 15;
    board[x][y] = dot;
  }

  int weight[9];

  for (i=0; i<9; i++)
  {
    if (results[i] == X_WINS && this_player == '1'
	|| results[i] == O_WINS && this_player == '0')
      weight[i] = moves[i];
    else if (results[i] == DRAW)
      weight[i] = 100 - moves[i];
    else if (results[i] != NO_MOVE)
      weight[i] = 200 - moves[i];
    else
      weight[i] = 1000 - moves[i];
  }

  int win_move = 0;
  for (i=1; i<9; i++)
    if (weight[win_move] > weight[i])
      win_move = i;

  note(this_player, win_move+'0');

  if (!quiet)
  {
    if (verbose)
    {
      x = win_move/3;
      y = win_move%3;

      board[x][y] = '*';

      printf(" %c | %c | %c  Player: %c\n",
	     board[0][0], board[0][1], board[0][2], this_player);
      printf("---+---+---\n");
      printf(" %c | %c | %c  Move: %d,%d\n",
	     board[1][0], board[1][1], board[1][2], x, y);
      printf("---+---+---\n");
      printf(" %c | %c | %c  Results: %s\n",
	     board[2][0], board[2][1], board[2][2], describe[results[win_move]]);

      board[x][y] = dot;

      printf("\n");
      for (i=0; i<9; i++)
      {
	printf(" (r=%d,m=%d,w=%4d)", results[i], moves[i], weight[i]);
	if (i % 3 == 2) printf("\n");
      }
      printf("\n");
    }
  }

  saved_results[board_index] = results[win_move] + 16*(moves[win_move]+1);
  return saved_results[board_index];
}

int
main(int argc, char **argv)
{
  memset(saved_results, UNKNOWN, sizeof(saved_results));

  if (argc > 1 && strcmp(argv[1], "-g") == 0)
  {
    for (int i=0; i<9; i++)
      board[i/3][i%3] = argv[2][i];
    verbose=1;
    quiet = 0;
    outf = stdout;
    figure_board('1', '0', 0);
    exit(0);
  }

  if (argc > 1 && strcmp(argv[1], "-q") == 0)
  {
    quiet = 1;
    outf = stdout;
    argc--;
    argv++;
  }
  if (argc > 1 && strcmp(argv[1], "-v") == 0)
  {
    verbose = 1;
    argc--;
    argv++;
  }

  if (argc > 1)
    outf = fopen(argv[1], "w");

  int a, b, c, d, e, f, g, h, i;

  for (a=0; a<3; a++)
  {
    board[0][0] = key[a];
    for (b=0; b<3; b++)
    {
      board[0][1] = key[b];
      for (c=0; c<3; c++)
      {
	board[0][2] = key[c];
	for (d=0; d<3; d++)
	{
	  board[1][0] = key[d];
	  for (e=0; e<3; e++)
	  {
	    board[1][1] = key[e];
	    for (f=0; f<3; f++)
	    {
	      board[1][2] = key[f];
	      for (g=0; g<3; g++)
	      {
		board[2][0] = key[g];
		for (h=0; h<3; h++)
		{
		  board[2][1] = key[h];
		  for (i=0; i<3; i++)
		  {
		    board[2][2] = key[i];
		    figure_board('1', '0', 0);
		    figure_board('0', '1', 0);
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }

  if (outf && outf != stdout)
    fclose(outf);

  if (!quiet)
    printf("\n%d games considered\n", count);

  return 0;
}
