#!/bin/perl
# -*- perl -*-
# Copyright 1997 DJ Delorie <dj@delorie.com>
# Permission granted to do whatever you want with these sources,
# provided they are not changed (except for headers and trailers, where noted)

use GDBM_File;

if (!$ARGV[0] && !$ENV{'SERVER_NAME'}) {
    &generate_database();
    exit 0;
}

print "Content-type: text/html\n\n";
&header("Tic Tac Toe");

tie (%INFO, "GDBM_File", "games", GDBM_READER, 0644)
    || die("cannot tie games: $!");

@winning_patterns = ("111......", "000......", 
		     "...111...", "...000...", 
		     "......111", "......000", 
		     "1..1..1..", "0..0..0..", 
		     ".1..1..1.", ".0..0..0.", 
		     "..1..1..1", "..0..0..0", 
		     "1...1...1", "0...0...0", 
		     "..1.1.1..", "..0.0.0.." );

$board = shift;
$board =~ tr/1O/1o/;

if ($board !~ /^[10-]{10}$/) {
    print "<h2 align=center>Hey, you tried to fake me out,\n";
    print "and it didn't work.</h2>\n";
    &trailer();
}

$computer = chop($board);

if ($computer eq "1") {
    $x = "computer";
    $o = "human";
    $human = "0";
} else {
    $x = "human";
    $o = "computer";
    $computer = "0";
    $human = "1";
}

&check_winners();

$move = $DB{$board.$computer};

if (! defined $move) {
    $win = "a";
    &show_board();
    print "<h2 align=center>This is an invalid board.</h2>\n";
    &trailer();
}

substr($board,$move,1) = $computer;

&check_winners();

&show_board();

&trailer();

#-----------------------------------------------------------------------------

sub check_winners {
    for $i (0..$#winning_patterns) {
	$p = $winning_patterns[$i];
	if ($board =~ /$p/) {
	    $win = $p;
	    &show_board();
	    print "<h2 align=center>";
	    print $win =~ /1/ ? $x : $o;
	    print " wins!</h2>\n\n";
	    &trailer();
	}
    }
    if ($board !~ /-/) {
	$win = "---------";
	&show_board();
	print "<h2 align=center>Draw!</h2>\n";
	&trailer();
    }
}

sub show_board {
    print "<center><p align=center>\n";
    print "<table border=0 cellspacing=0 cellpadding=0>\n";

    print "<tr>"; &cell(0); &vbar; &cell(1); &vbar; &cell(2); print "</tr>";
    &hbar();
    print "<tr>"; &cell(3); &vbar; &cell(4); &vbar; &cell(5); print "</tr>";
    &hbar();
    print "<tr>"; &cell(6); &vbar; &cell(7); &vbar; &cell(8); print "</tr>";
    print "</table>\n\n";
    print "</p></center>\n\n";
}

sub cell {
    $c = substr($board,$_[0],1);
    $w = substr($win,$_[0],1);
    if ($c eq '1') {
	if ($w eq '1') {
	    $gif = $x."-win";
	    $alt = "1";
	} else {
	    $gif = $x;
	    $alt = "1";
	}
    } elsif ($c eq '0') {
	if ($w eq '0') {
	    $gif = $o."-win";
	    $alt = "0";
	} else {
	    $gif = $o;
	    $alt = "0";
	}
    } else {
	$gif = "blank";
	$alt = ".";
    }

    $bd2 = $board;
    substr($bd2, $_[0], 1) = $human;
    print "<td>";
    print "<a href=\"ttt.cgi?$bd2$computer\">" unless $win;
    print "<img src=\"$gif.gif\" width=50 height=50 alt=\"$alt\" border=0>";
    print "</a>" unless $win;
    print "</td>\n";
}

sub vbar {
    print "<td><img src=\"vbar.gif\" alt=\"\" width=3 height=50></td>\n";
}

sub hbar {
    print "<tr><td colspan=5><img src=\"hbar.gif\" alt=\"\" width=156 height=3></td></tr>\n";
}

#-----------------------------------------------------------------------------

sub header {
    # You may change this to generate your header
    print `header $_[0]`;
    #print "<title>$_[0]</title><h1 align=center>$_[0]</h1>\n\n";
}

sub trailer {
    print "<p align=center> <a\n";
    print "href=\"http://www.delorie.com/game-room/tictactoe/\">Written by\n";
    print "Delorie Software</a></p>\n\n";

    # You may change this to generate your trailer
    print `trailer`;
    #print "<p><hr>\n";

    exit 0;
}

#-----------------------------------------------------------------------------

sub generate_database {
    $| = 1;
    print "generating database...\n";
    unlink "games.tmp.pag", "games.tmp.dir";
    open(G, "./rungames -q |");
    tie (%DB, "GDBM_File", "games.tmp", GDBM_WRCREAT, 0644)
	|| die("cannot tie games: $!");

    $old_percent = -1;
    $count = 0;
    while (<G>) {
	($board, $move, $result) = split(' ');
	$DB{$board} = $move;
	$count++;
	$new_percent = int($count * 100 / 39366);
	if ($new_percent != $old_percent) {
	    printf("\r%3d%% ", $new_percent);
	    $old_percent = $new_percent;
	}
    }

    untie %DB;
    close(G);
    print "\n\n";
    system "ls games*";
    rename("games.tmp", "games");
}
